#!/usr/bin/env python
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Bit9 + Carbon Black
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from cbapi.event import on_event, registry
from cbapi.example_helpers import get_cb_response_object, build_cli_parser
import re
from cbapi.response import sensor_events, event
import sys
import time

autoruns = """\\registry\\machine\\system\\currentcontrolset\\control\\session manager\\bootexecute(.*)
\\registry\\machine\\system\\currentcontrolset\\services(.*)
\\registry\\machine\\system\\currentcontrolset\\services(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\runservicesonce(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\runservices(.*)
\\registry\\machine\\software\\microsoft\\windows nt\\currentversion\\winlogon\\notify(.*)
\\registry\\machine\\software\\microsoft\\windows nt\\currentversion\\winlogon\\userinit(.*)
\\registry\\machine\\software\\microsoft\\windows nt\\currentversion\\winlogon\\shell(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\shellserviceobjectdelayload(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\runonce(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\runonceex(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\run(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\policies\\explorer\\run(.*)
\\registry\\machine\\software\\microsoft\\windows nt\\currentversion\\windows(.*)
\\registry\\machine\\software\\microsoft\\windows\\currentversion\\explorer\\sharedtaskscheduler(.*)
\\registry\\machine\\software\\microsoft\\windows nt\\currentversion\\windows\\appinit_dlls(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows\\currentversion\\runservicesonce(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows\\currentversion\\runservices(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows nt\\currentversion\\winlogon\\shell(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows\\currentversion\\run(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows\\currentversion\\runonce(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows\\currentversion\\policies\\explorer\\run(.*)
\\registry\\user\\(.*)\\software\\microsoft\\windows nt\\currentversion\\windows\\load(.*)"""

autoruns_regex = re.compile("|".join(autoruns.replace("\\", "\\\\").split("\n")))


class GetRegistryValue(object):
    def __init__(self, registry_key):
        self.registry_key = registry_key

    def run(self, session):
        reg_info = session.get_registry_value(self.registry_key)
        return time.time(), session.sensor_id, self.registry_key, reg_info["value_data"]


@on_event("ingress.event.regmod")
def process_callback(cb, event_type, event_data):
    x = sensor_events.CbEventMsg()
    x.ParseFromString(event_data)

    regmod_path = str(x.regmod.utf8_regpath)

    if autoruns_regex.match(regmod_path):
        regmod_path = regmod_path.replace("\\registry\\machine\\", "HKLM\\")
        regmod_path = regmod_path.replace("\\registry\\user\\", "HKEY_USERS\\")
        regmod_path = regmod_path.strip()
        job = GetRegistryValue(regmod_path)
        registry_job = cb.live_response.submit_job(job.run, x.env.endpoint.SensorId)
        registry_job.add_done_callback(print_result)


def print_result(registry_job):
    try:
        timestamp, sensor_id, registry_key, registry_value = registry_job.result()
    except Exception:
        print("Error encountered when pulling registry key: {0}".format(registry_job.exception()))
    else:
        print("Got result for sensor ID {0} registry key {1}: value is {2}".format(sensor_id, registry_key,
                                                                                   registry_value))


def main():
    parser = build_cli_parser("Get value from any new regmods to autorun keys")
    args = parser.parse_args()
    cb = get_cb_response_object(args)

    # event_source = event.RabbitMQEventSource(cb)
    event_source = event.RabbitMQEventSource(cb)
    event_source.start()

    print("Listening on the event bus for regmods")

    # print_thread = ResultPrinter(cb)
    # print_thread.start()

    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        event_source.stop()
        print("Exiting event loop")

    print("Encountered the following exceptions during processing:")
    for error in registry.errors:
        print(error["exception"])


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from cbapi.protection import Connector, Notification
from cbapi.example_helpers import get_cb_protection_object, build_cli_parser
import logging
import time


log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


def process_request(api, i):
    # make sure we have the latest information
    i.refresh()

    log.info("Processing hash {0:s}".format(i.fileHash))

    if i.uploaded:
        log.info("I have a binary! let's download it")
        open(i.fileHash+".dat", "wb").write(i.file.read())

        # create a notification
        n = i.create_notification(data={"product": "EmptyAnalysis", "malwareName": "TEST MALWARE",
                                        "analysisResult": Notification.ResultMalicious, "severity": "high",
                                        "type": "whoa"})
        n.save()
    else:
        # wait till next time
        log.info("Waiting until hash {0:s} is uploaded".format(i.fileHash))


if __name__ == '__main__':
    parser = build_cli_parser("EmptyAnalysis: example connector for Carbon Black Enterprise Protection")
    args = parser.parse_args()

    api = get_cb_protection_object(args)
    vt_connector = api.create(Connector, data={"name": "EmptyAnalysis", "analysisName": "EmptyAnalysis",
                                               "connectorVersion": "1.0", "canAnalyze": True, "analysisEnabled": True,
                                               "enabled": True})
    vt_connector.save()

    connector_id = vt_connector.id
    log.info("Connector: {0:s}".format(str(vt_connector)))

    log.info("Starting processing loop")
    while True:
        for i in vt_connector.pendingAnalyses:
            process_request(api, i)
        time.sleep(10)



import logging
import threading
from cbapi.six.moves.queue import Queue
from collections import defaultdict
from concurrent.futures import _base

from cbapi.errors import TimeoutError
from cbapi.live_response_api import CbLRManagerBase, CbLRSessionBase, poll_status


OS_LIVE_RESPONSE_ENUM = {
    "WINDOWS": 1,
    "LINUX": 2,
    "MAC": 4
}

log = logging.getLogger(__name__)


class LiveResponseSession(CbLRSessionBase):
    def __init__(self, cblr_manager, session_id, sensor_id, session_data=None):
        super(LiveResponseSession, self).__init__(cblr_manager, session_id, sensor_id, session_data=session_data)
        from cbapi.psc.defense.models import Device
        device_info = self._cb.select(Device, self.sensor_id)
        self.os_type = OS_LIVE_RESPONSE_ENUM.get(device_info.deviceType, None)


class WorkItem(object):
    def __init__(self, fn, sensor_id):
        from cbapi.psc.defense.models import Device
        self.fn = fn
        if isinstance(sensor_id, Device):
            self.sensor_id = sensor_id.deviceId
        else:
            self.sensor_id = int(sensor_id)

        self.future = _base.Future()


class CompletionNotification(object):
    def __init__(self, sensor_id):
        self.sensor_id = sensor_id


class WorkerStatus(object):
    def __init__(self, sensor_id, status="ready", exception=None):
        self.sensor_id = sensor_id
        self.status = status
        self.exception = exception


class JobWorker(threading.Thread):
    def __init__(self, cb, sensor_id, result_queue):
        super(JobWorker, self).__init__()
        self.cb = cb
        self.sensor_id = sensor_id
        self.job_queue = Queue()
        self.lr_session = None
        self.result_queue = result_queue

    def run(self):
        try:
            self.lr_session = self.cb.live_response.request_session(self.sensor_id)
            self.result_queue.put(WorkerStatus(self.sensor_id, status="ready"))

            while True:
                work_item = self.job_queue.get(block=True)
                if not work_item:
                    self.job_queue.task_done()
                    return

                self.run_job(work_item)
                self.result_queue.put(CompletionNotification(self.sensor_id))
                self.job_queue.task_done()
        except Exception as e:
            self.result_queue.put(WorkerStatus(self.sensor_id, status="error", exception=e))
        finally:
            if self.lr_session:
                self.lr_session.close()
            self.result_queue.put(WorkerStatus(self.sensor_id, status="exiting"))

    def run_job(self, work_item):
        try:
            work_item.future.set_result(work_item.fn(self.lr_session))
        except Exception as e:
            work_item.future.set_exception(e)


class LiveResponseSessionManager(CbLRManagerBase):
    cblr_base = "/integrationServices/v3/cblr"
    cblr_session_cls = LiveResponseSession

    def submit_job(self, job, sensor):
        if self._job_scheduler is None:
            # spawn the scheduler thread
            self._job_scheduler = LiveResponseJobScheduler(self._cb)
            self._job_scheduler.start()

        work_item = WorkItem(job, sensor)
        self._job_scheduler.submit_job(work_item)
        return work_item.future

    def _get_or_create_session(self, sensor_id):
        session_id = self._create_session(sensor_id)

        try:
            res = poll_status(self._cb, "{cblr_base}/session/{0}".format(session_id, cblr_base=self.cblr_base),
                              desired_status="ACTIVE", delay=1, timeout=360)
        except Exception:
            # "close" the session, otherwise it will stay in a pending state
            self._close_session(session_id)

            # the Cb server will return a 404 if we don't establish a session in time, so convert this to a "timeout"
            raise TimeoutError(uri="{cblr_base}/session/{0}".format(session_id, cblr_base=self.cblr_base),
                               message="Could not establish session with sensor {0}".format(sensor_id),
                               error_code=404)
        else:
            return session_id, res

    def _close_session(self, session_id):
        try:
            self._cb.put_object("{cblr_base}/session".format(session_id, cblr_base=self.cblr_base),
                                {"session_id": session_id, "status": "CLOSE"})
        except Exception:
            pass

    def _create_session(self, sensor_id):
        response = self._cb.post_object("{cblr_base}/session/{0}".format(sensor_id, cblr_base=self.cblr_base),
                                        {"sensor_id": sensor_id}).json()
        session_id = response["id"]
        return session_id


class LiveResponseJobScheduler(threading.Thread):
    daemon = True

    def __init__(self, cb, max_workers=10):
        super(LiveResponseJobScheduler, self).__init__()
        self._cb = cb
        self._job_workers = {}
        self._idle_workers = set()
        self._unscheduled_jobs = defaultdict(list)
        self._max_workers = max_workers
        self.schedule_queue = Queue()

    def run(self):
        log.debug("Starting Live Response Job Scheduler")

        while True:
            log.debug("Waiting for item on Scheduler Queue")
            item = self.schedule_queue.get(block=True)
            log.debug("Got item: {0}".format(item))
            if isinstance(item, WorkItem):
                # new WorkItem available
                self._unscheduled_jobs[item.sensor_id].append(item)
            elif isinstance(item, CompletionNotification):
                # job completed
                self._idle_workers.add(item.sensor_id)
            elif isinstance(item, WorkerStatus):
                if item.status == "error":
                    log.error("Error encountered by JobWorker[{0}]: {1}".format(item.sensor_id,
                                                                                item.exception))
                elif item.status == "exiting":
                    log.debug("JobWorker[{0}] has exited, waiting...".format(item.sensor_id))
                    self._job_workers[item.sensor_id].join()
                    log.debug("JobWorker[{0}] deleted".format(item.sensor_id))
                    del self._job_workers[item.sensor_id]
                    try:
                        self._idle_workers.remove(item.sensor_id)
                    except KeyError:
                        pass
                elif item.status == "ready":
                    log.debug("JobWorker[{0}] now ready to accept jobs, session established".format(item.sensor_id))
                    self._idle_workers.add(item.sensor_id)
                else:
                    log.debug("Unknown status from JobWorker[{0}]: {1}".format(item.sensor_id, item.status))
            else:
                log.debug("Received unknown item on the scheduler Queue, exiting")
                # exiting the scheduler if we get None
                # TODO: wait for all worker threads to exit
                return

            self._schedule_jobs()

    def _schedule_jobs(self):
        log.debug("Entering scheduler")

        # First, see if there are new jobs to schedule on idle workers.
        self._schedule_existing_workers()

        # If we have jobs scheduled to run on sensors with no current associated worker, let's spawn new ones.
        if set(self._unscheduled_jobs.keys()) - self._idle_workers:
            self._cleanup_idle_workers()
            self._spawn_new_workers()
            self._schedule_existing_workers()

    def _cleanup_idle_workers(self, max=None):
        if not max:
            max = self._max_workers

        for sensor in list(self._idle_workers)[:max]:
            log.debug("asking worker for sensor id {0} to exit".format(sensor))
            self._job_workers[sensor].job_queue.put(None)

    def _schedule_existing_workers(self):
        log.debug("There are idle workers for sensor ids {0}".format(self._idle_workers))

        intersection = self._idle_workers.intersection(set(self._unscheduled_jobs.keys()))

        log.debug("{0} jobs ready to execute in existing execution slots".format(len(intersection)))

        for sensor in intersection:
            item = self._unscheduled_jobs[sensor].pop(0)
            self._job_workers[sensor].job_queue.put(item)
            self._idle_workers.remove(item.sensor_id)

        self._cleanup_unscheduled_jobs()

    def _cleanup_unscheduled_jobs(self):
        marked_for_deletion = []
        for k in self._unscheduled_jobs.keys():
            if len(self._unscheduled_jobs[k]) == 0:
                marked_for_deletion.append(k)

        for k in marked_for_deletion:
            del self._unscheduled_jobs[k]

    def submit_job(self, work_item):
        self.schedule_queue.put(work_item)

    def _spawn_new_workers(self):
        from cbapi.psc.defense.models import Device
        if len(self._job_workers) >= self._max_workers:
            return

        schedule_max = self._max_workers - len(self._job_workers)
        '''
        sensors = [s for s in self._cb.select(Device) if s.deviceId in self._unscheduled_jobs
                    and s.deviceId not in self._job_workers
                    and "AVAILABLE" in s.sensorStates]
        '''
        log.debug("spawning new workers to handle unscheduled jobs: {0}".format(self._unscheduled_jobs))
        sensors = [s for s in self._cb.select(Device) if s.deviceId in self._unscheduled_jobs
                   and s.deviceId not in self._job_workers]
        sensors_to_schedule = sensors[:schedule_max]
        log.debug("Spawning new workers to handle these sensors: {0}".format(sensors_to_schedule))
        for sensor in sensors_to_schedule:
            log.debug("Spawning new JobWorker for sensor id {0}".format(sensor.deviceId))
            self._job_workers[sensor.deviceId] = JobWorker(self._cb, sensor.deviceId, self.schedule_queue)
            self._job_workers[sensor.deviceId].start()



class GetFileJob(object):
    def __init__(self, file_name):
        self._file_name = file_name

    def run(self, session):
        return session.get_file(self._file_name)


def getjob():
    return GetFileJob("c:\\test.txt")



import sys
from distutils.version import LooseVersion

from cbapi.response.models import StoragePartition
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import logging

log = logging.getLogger(__name__)


def list_partitions(cb, parser, args):
    for partition in cb.select(StoragePartition):
        print('{0} Storage Partition "{1}":'.format(partition.status.upper(), partition.name))
        for partition_info in partition.info:
            print("  {0:15s}: {1}".format(partition_info, partition.info[partition_info]))


def create_partition(cb, parser, args):
    cb.create_new_partition()
    print("Successfully created new writer partition")


def delete_partition(cb, parser, args):
    # select the partition by name (which same as the ID)
    try:
        partition = cb.select(StoragePartition, args.name)
        partition.delete()
    except Exception as e:
        print("Error deleting partition {0}: {1}".format(args.name, str(e)))
    else:
        print("Successfully deleted partition {0}".format(args.name))


def mount_partition(cb, parser, args):
    # select the partition by name (which same as the ID)
    try:
        partition = cb.select(StoragePartition, args.name)
        partition.mount()
    except Exception as e:
        print("Error mounting partition {0}: {1}".format(args.name, str(e)))
    else:
        print("Successfully mounted partition {0}".format(args.name))


def unmount_partition(cb, parser, args):
    # select the partition by name (which same as the ID)
    try:
        partition = cb.select(StoragePartition, args.name)
        partition.unmount()
    except Exception as e:
        print("Error unmounting partition {0}: {1}".format(args.name, str(e)))
    else:
        print("Successfully unmounted partition {0}".format(args.name))


def main():
    parser = build_cli_parser()
    commands = parser.add_subparsers(help="Storage Partition commands", dest="command_name")

    commands.add_parser("list", help="List all storage partitions")

    commands.add_parser("create", help="Create new active writer partition")

    del_command = commands.add_parser("delete", help="Delete partition")
    del_command.add_argument("-N", "--name", help="Name of partition to delete.", required=True)

    mount_command = commands.add_parser("mount", help="Mount partition")
    mount_command.add_argument("-N", "--name", help="Name of partition to mount.", required=True)

    unmount_command = commands.add_parser("unmount", help="Unmount partition")
    unmount_command.add_argument("-N", "--name", help="Name of partition to unmount.", required=True)

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    if cb.cb_server_version < LooseVersion("6.1.0"):
        parser.error("This script can only work with server versions >= 6.1.0; {0} is running {1}"
                     .format(cb.url, cb.cb_server_version))
        return 1

    if args.command_name == "list":
        return list_partitions(cb, parser, args)
    elif args.command_name == "create":
        return create_partition(cb, parser, args)
    elif args.command_name == "delete":
        return delete_partition(cb, parser, args)
    elif args.command_name == "mount":
        return mount_partition(cb, parser, args)
    elif args.command_name == "unmount":
        return unmount_partition(cb, parser, args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_defense_object
from cbapi.psc.defense import Device


def main():
    parser = build_cli_parser("List devices")
    device_options = parser.add_mutually_exclusive_group(required=False)
    device_options.add_argument("-i", "--id", type=int, help="Device ID of sensor")
    device_options.add_argument("-n", "--hostname", help="Hostname")

    args = parser.parse_args()
    cb = get_cb_defense_object(args)

    if args.id:
        devices = [cb.select(Device, args.id)]
    elif args.hostname:
        devices = list(cb.select(Device).where("hostNameExact:{0}".format(args.hostname)))
    else:
        devices = list(cb.select(Device))

    print("{0:9} {1:40}{2:18}{3}".format("ID", "Hostname", "IP Address", "Last Checkin Time"))
    for device in devices:
        print("{0:9} {1:40s}{2:18s}{3}".format(device.deviceId, device.name or "None",
                                               device.lastInternalIpAddress or "Unknown", device.lastContact))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys
from cbapi.example_helpers import build_cli_parser, get_cb_psc_object
from cbapi.psc.models import BaseAlert
from helpers.alertsv6 import setup_parser_with_basic_criteria, load_basic_criteria


def main():
    parser = build_cli_parser("List alerts")
    setup_parser_with_basic_criteria(parser)
    parser.add_argument("-S", "--sort_by", help="Field to sort the output by")
    parser.add_argument("-R", "--reverse", action="store_true", help="Reverse order of sort")

    args = parser.parse_args()
    cb = get_cb_psc_object(args)

    query = cb.select(BaseAlert)
    load_basic_criteria(query, args)
    if args.sort_by:
        direction = "DESC" if args.reverse else "ASC"
        query = query.sort_by(args.sort_by, direction)

    alerts = list(query)
    print("{0:40} {1:40s} {2:40s} {3}".format("ID", "Hostname", "Threat ID", "Last Updated"))
    for alert in alerts:
        print("{0:40} {1:40s} {2:40s} {3}".format(alert.id, alert.device_name or "None",
                                                  alert.threat_id or "Unknown",
                                                  alert.last_update_time))


if __name__ == "__main__":
    sys.exit(main())



import enum
import logging


class IOC_v2():
    """Models an indicator of compromise detected during an analysis.

    Every IOC belongs to an AnalysisResult.
    """

    def __init__(self, analysis, match_type, values, field, link):
        self.id = analysis
        self.match_type = match_type
        self.values = values
        self.field = field
        self.link = link

    class MatchType(str, enum.Enum):
        """
        Represents the valid matching strategies for an IOC.
        """

        Equality: str = "equality"
        Regex: str = "regex"
        Query: str = "query"

    def as_dict(self):
        return {
            "id": str(self.id),
            "match_type": self.match_type,
            "values": list(self.values),
            "field": self.field,
            "link": self.link,
        }


class AnalysisResult():
    """Models the result of an analysis performed by a connector."""

    def __init__(self, analysis_name, scan_time, score, title, description):
        self.id = str(analysis_name)
        self.timestamp = scan_time
        self.title = title
        self.description = description
        self.severity = score
        self.iocs = []
        self.iocs_v2 = []
        self.link = None
        self.tags = None
        self.visibility = None
        self.connector_name = "STIX_TAXII"

    def attach_ioc_v2(self, *, match_type=IOC_v2.MatchType.Equality, values, field, link):
        self.iocs_v2.append(IOC_v2(analysis=self.id, match_type=match_type, values=values, field=field, link=link))

    def normalize(self):
        """Normalizes this result to make it palatable for the CbTH backend."""

        if self.severity <= 0 or self.severity > 10:
            logging.warning("normalizing OOB score: {}".format(self.severity))
            self.severity = max(1, min(self.severity, 10))
            # NOTE: min 1 and not 0
            # else err 400 from cbapi: Report severity must be between 1 & 10
        return self

    def as_dict(self):
        return {"IOCs_v2": [ioc_v2.as_dict() for ioc_v2 in self.iocs_v2], **super().as_dict()}

    def as_dict_full(self):
        return {
            "id": self.id,
            "timestamp": self.timestamp,
            "title": self.title,
            "description": self.description,
            "severity": self.severity,
            "iocs_v2": [iocv2.as_dict() for iocv2 in self.iocs_v2]
        }



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_object
from cbapi.psc.threathunter import Process, Binary
from cbapi.errors import ObjectNotFoundError


def main():
    parser = build_cli_parser("Query processes")
    parser.add_argument("-q", type=str, help="process query", default="process_name:notepad.exe")
    parser.add_argument("-n", type=int, help="only output N processes", default=None)
    parser.add_argument("-b", action="store_true", help="show binary information", default=False)

    args = parser.parse_args()
    cb = get_cb_threathunter_object(args)

    processes = cb.select(Process).where(args.q)

    if args.n:
        processes = processes[0:args.n]

    for process in processes:
        print("Process: {}".format(process.process_name))
        print("\tPIDs: {}".format(process.process_pids))
        print("\tSHA256: {}".format(process.process_sha256))
        print("\tGUID: {}".format(process.process_guid))

        if args.b:
            try:
                binary = cb.select(Binary, process.process_sha256)
                print(binary)
                print(binary.summary)
            except ObjectNotFoundError:
                print("No binary found for process with hash: {}".format(process.process_sha256))


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_object
from cbapi.psc.threathunter import Process


def main():
    parser = build_cli_parser("Search processes")
    parser.add_argument("-q", type=str, help="process query", default="process_name:notepad.exe")
    parser.add_argument("-f", help="show full objects", action="store_true", default=False)
    parser.add_argument("-n", type=int, help="only output N processes", default=None)
    parser.add_argument("-e", help="show events for query results", action="store_true", default=False)
    parser.add_argument("-c", help="show children for query results", action="store_true", default=False)
    parser.add_argument("-p", help="show parents for query results", action="store_true", default=False)
    parser.add_argument("-t", help="show tree for query results", action="store_true", default=False)
    parser.add_argument("-S", type=str, help="sory by this field", required=False)
    parser.add_argument("-D", help="return results in descending order", action="store_true")

    args = parser.parse_args()
    cb = get_cb_threathunter_object(args)

    processes = cb.select(Process).where(args.q)

    direction = "ASC"
    if args.D:
        direction = "DESC"

    if args.S:
        processes.sort_by(args.S, direction=direction)

    print("Number of processes: {}".format(len(processes)))

    if args.n:
        processes = processes[0:args.n]

    for process in processes:
        if args.f:
            print(process)
        else:
            print("{} ({}): {}".format(process.process_name, process.process_guid, process.process_sha256))

        if args.e:
            print("=========== events ===========")
            for event in process.events():
                if args.f:
                    print(event)
                else:
                    print("\t{}".format(event.event_type))

        if args.c:
            print("========== children ==========")
            for child in process.children:
                if args.f:
                    print(child)
                else:
                    print("\t{}: {}".format(child.process_name, child.process_sha256))

        if args.p:
            print("========== parents ==========")
            for parent in process.parents:
                if args.f:
                    print(parent)
                else:
                    print("\t{}: {}".format(parent.process_name, parent.process_sha256))

        if args.t:
            print("=========== tree =============")
            tree = process.tree()
            print(tree)
            print(tree.nodes)

        print("===========================")


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from cbapi.protection import Computer
from cbapi.example_helpers import get_cb_protection_object, build_cli_parser
import sys


def main():
    parser = build_cli_parser("Approve files")
    parser.add_argument("--policy", "-p", help="Policy name to use", required=True)
    parser.add_argument("--threshold", "-t", help="Prevalence threshold", default=10, type=int)
    args = parser.parse_args()

    api = get_cb_protection_object(args)

    comps = api.select(Computer).where("policyName:{0:s}".format(args.policy)).and_("deleted:false")
    for c in comps:
        changed_files = []
        files = c.fileInstances.where("localState:1")
        for f in files:
            if f.fileCatalog.prevalence >= args.threshold:
                f.localState = 2
                f.save()
                changed_files.append(f)

        if len(changed_files) > 0:
            print("Locally approved {0:d} files on computer {1:s}".format(len(changed_files), c.name))


if __name__ == '__main__':
    sys.exit(main())



from __future__ import absolute_import

import cbapi.six

__title__ = 'cbapi'
__author__ = 'Carbon Black Developer Network'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018-2019 VMware Carbon Black'
__version__ = '1.6.2'

# New API as of cbapi 0.9.0
from cbapi.response.rest_api import CbEnterpriseResponseAPI, CbResponseAPI
from cbapi.protection.rest_api import CbEnterpriseProtectionAPI, CbProtectionAPI
from cbapi.psc import CbPSCBaseAPI
from cbapi.psc.defense import CbDefenseAPI
from cbapi.psc.threathunter import CbThreatHunterAPI
from cbapi.psc.livequery import CbLiveQueryAPI

# for compatibility with Cb Defense code from cbapi < 1.4.0
import cbapi.psc.defense as defense



import sys

from cbapi.example_helpers import build_cli_parser, get_cb_livequery_object
from cbapi.psc.livequery.models import Result


def main():
    parser = build_cli_parser("Search the device summaries of a LiveQuery run")
    parser.add_argument("-i", "--id", type=str, required=True, help="Run ID")
    parser.add_argument("-q", "--query", type=str, required=False, help="Search query")

    parser.add_argument(
        "--device_ids",
        nargs="+",
        type=int,
        required=False,
        help="Device IDs to filter on",
    )
    parser.add_argument(
        "--device_names",
        nargs="+",
        type=int,
        required=False,
        help="Device names to filter on",
    )
    parser.add_argument(
        "--policy_ids",
        nargs="+",
        type=int,
        required=False,
        help="Policy IDs to filter on",
    )
    parser.add_argument(
        "--policy_names",
        nargs="+",
        type=int,
        required=False,
        help="Policy names to filter on",
    )
    parser.add_argument(
        "--statuses",
        nargs="+",
        type=str,
        required=False,
        help="Statuses to filter on",
    )
    parser.add_argument(
        "-S", "--sort_by", type=str, help="sort by this field", required=False
    )
    parser.add_argument(
        "-D",
        "--descending_results",
        help="return results in descending order",
        action="store_true",
    )

    args = parser.parse_args()
    cb = get_cb_livequery_object(args)

    results = cb.select(Result).run_id(args.id)
    result = results.first()
    if result is None:
        print("ERROR: No results.")
        return 1

    summaries = result.query_device_summaries()
    if args.query:
        summaries.where(args.query)
    if args.device_ids:
        summaries.criteria(device_id=args.device_ids)
    if args.device_names:
        summaries.criteria(device_name=args.device_names)
    if args.policy_ids:
        summaries.criteria(policy_id=args.policy_ids)
    if args.policy_names:
        summaries.criteria(policy_name=args.policy_names)
    if args.statuses:
        summaries.criteria(status=args.statuses)
    if args.sort_by:
        dir = "DESC" if args.descending_results else "ASC"
        summaries.sort_by(args.sort_by, direction=dir)

    for summary in summaries:
        print(summary)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys

from cbapi.example_helpers import build_cli_parser, get_cb_threathunter_object
from cbapi.psc.threathunter import Process
import json
import csv


def main():
    parser = build_cli_parser("Query processes")
    parser.add_argument("-p", type=str, help="process guid", default=None)
    parser.add_argument("-q", type=str, help="query string", default=None)
    parser.add_argument("-s", type=bool, help="silent mode", default=False)
    parser.add_argument("-n", type=int, help="only output N events", default=None)
    parser.add_argument("-f", type=str, help="output file name", default=None)
    parser.add_argument("-of", type=str, help="output file format: csv or json", default="json")

    args = parser.parse_args()
    cb = get_cb_threathunter_object(args)

    if not args.p and not args.q:
        print("Error: Missing Process GUID to search for events with")
        sys.exit(1)

    if args.q:
        processes = cb.select(Process).where(args.q)
    else:
        processes = cb.select(Process).where(process_guid=args.p)

    if args.n:
        processes = [p for p in processes[0:args.n]]

    if not args.s:
        for process in processes:
            print("Process: {}".format(process.process_name))
            print("\tPIDs: {}".format(process.process_pids))
            print("\tSHA256: {}".format(process.process_sha256))
            print("\tGUID: {}".format(process.process_guid))

    if args.f is not None:
        if args.of == "json":
            with open(args.f, 'w') as outfile:
                for p in processes:
                    json.dump(p.original_document, outfile)
        else:
            with open(args.f, 'w') as outfile:
                csvwriter = csv.writer(outfile)
                for p in processes:
                    csvwriter.writerow(p.original_document)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys
from cbapi.response.models import SensorGroup
from cbapi.errors import ObjectNotFoundError
from cbapi.example_helpers import build_cli_parser, get_cb_response_object


def main():
    parser = build_cli_parser()
    parser.add_argument("--filename", "-f", action="store", default=None, dest="filename",
                        help="Filename to save the installer package to", required=True)
    parser.add_argument("--sensor-group", "-g", action="store", default="1", dest="group",
                        help="Sensor group name or ID of the group to download an installer for")
    parser.add_argument("--installer-type", "-t", action="store", default="windows/exe", dest="type",
                        help="Installer type; must be one of windows/exe, windows/msi, linux, osx")

    args = parser.parse_args()
    cb = get_cb_response_object(args)

    try:
        sensor_group_id = int(args.group)
        sensor_group = cb.select(SensorGroup, sensor_group_id, force_init=True)
    except (ValueError, ObjectNotFoundError):
        sensor_group = cb.select(SensorGroup).where('name:{0:s}'.format(args.group)).one()
    except Exception:
        print("Could not find sensor group via id or name ({0:s})".format(args.group))
        return 1

    # download the installer package
    #
    print("-> Downloading {0:s} installer for group {1:s} to file {2:s}...".format(args.type, sensor_group.name,
                                                                                   args.filename))
    try:
        open(args.filename, 'wb').write(sensor_group.get_installer(args.type))
    except ObjectNotFoundError:
        print("** Could not find an installer for {0:s}".format(args.type))
    except IOError:
        print("** Error writing to file {0:s}".format(args.filename))
    except Exception as e:
        print("** Unknown exception: {0:s}".format(str(e)))
    else:
        print("-> Download complete")


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#

import sys

import json
import logging

from cbapi.errors import ServerError
from cbapi.example_helpers import build_cli_parser, get_cb_defense_object
from cbapi.psc.defense import Policy

log = logging.getLogger(__name__)


def get_policy_by_name_or_id(cb, policy_id=None, name=None, return_all_if_none=False):
    policies = []

    try:
        if policy_id:
            if isinstance(policy_id, list):
                attempted_to_find = "IDs of {0}".format(", ".join([str(pid) for pid in policy_id]))
                policies = [p for p in cb.select(Policy) if p.id in policy_id]
            else:
                attempted_to_find = "ID of {0}".format(policy_id)
                policies = [cb.select(Policy, policy_id, force_init=True)]
        elif name:
            if isinstance(name, list):
                attempted_to_find = "names of {0}".format(", ".join(name))
                policies = [p for p in cb.select(Policy) if p.name in name]
            else:
                attempted_to_find = "name {0}".format(name)
                policies = [p for p in cb.select(Policy) if p.name == name]
        elif return_all_if_none:
            attempted_to_find = "all policies"
            policies = list(cb.select(Policy))
    except Exception as e:
        print("Could not find policy with {0}: {1}".format(attempted_to_find, str(e)))

    return policies


def list_policies(cb, parser, args):
    for p in cb.select(Policy):
        print(u"Policy id {0}: {1} {2}".format(p.id, p.name, "({0})".format(p.description) if p.description else ""))
        print("Rules:")
        for r in p.rules.values():
            print("  {0}: {1} when {2} {3} is {4}".format(r.get('id'), r.get("action"),
                                                          r.get("application", {}).get("type"),
                                                          r.get("application", {}).get("value"), r.get("operation")))


def import_policy(cb, parser, args):
    p = cb.create(Policy)

    p.policy = json.load(open(args.policyfile, "r"))
    p.description = args.description
    p.name = args.name
    p.priorityLevel = args.prioritylevel
    p.version = 2

    try:
        p.save()
    except ServerError as se:
        print("Could not add policy: {0}".format(str(se)))
    except Exception as e:
        print("Could not add policy: {0}".format(str(e)))
    else:
        print("Added policy. New policy ID is {0}".format(p.id))


def delete_policy(cb, parser, args):
    policies = get_policy_by_name_or_id(cb, args.id, args.name)
    if len(policies) == 0:
        return

    num_matching_policies = len(policies)
    if num_matching_policies > 1 and not args.force:
        print("{0:d} policies match and --force not specified. No action taken.".format(num_matching_policies))
        return

    for p in policies:
        try:
            p.delete()
        except Exception as e:
            print("Could not delete policy: {0}".format(str(e)))
        else:
            print("Deleted policy id {0} with name {1}".format(p.id, p.name))


def export_policy(cb, parser, args):
    policies = get_policy_by_name_or_id(cb, args.id, args.name, return_all_if_none=True)

    for p in policies:
        json.dump(p.policy, open("policy-{0}.json".format(p.id), "w"), indent=2)
        print("Wrote policy {0} {1} to file policy-{0}.json".format(p.id, p.name))


def add_rule(cb, parser, args):
    policies = get_policy_by_name_or_id(cb, args.id, args.name)

    num_matching_policies = len(policies)
    if num_matching_policies < 1:
        print("No policies match. No action taken.".format(num_matching_policies))

    for policy in policies:
        policy.add_rule(json.load(open(args.rulefile, "r")))
        print("Added rule from {0} to policy {1}.".format(args.rulefile, policy.name))


def del_rule(cb, parser, args):
    policies = get_policy_by_name_or_id(cb, args.id, args.name)

    num_matching_policies = len(policies)
    if num_matching_policies != 1:
        print("{0:d} policies match. No action taken.".format(num_matching_policies))

    policy = policies[0]
    policy.delete_rule(args.ruleid)

    print("Removed rule id {0} from policy {1}.".format(args.ruleid, policy.name))


def replace_rule(cb, parser, args):
    policies = get_policy_by_name_or_id(cb, args.id, args.name)

    num_matching_policies = len(policies)
    if num_matching_policies != 1:
        print("{0:d} policies match. No action taken.".format(num_matching_policies))

    policy = policies[0]
    policy.replace_rule(args.ruleid, json.load(open(args.rulefile, "r")))

    print("Replaced rule id {0} from policy {1} with rule from file {2}.".format(args.ruleid, policy.name,
                                                                                 args.rulefile))


def main():
    parser = build_cli_parser("Policy operations")
    commands = parser.add_subparsers(help="Policy commands", dest="command_name")

    commands.add_parser("list", help="List all configured policies")

    import_policy_command = commands.add_parser("import", help="Import policy from JSON file")
    import_policy_command.add_argument("-N", "--name", help="Name of new policy", required=True)
    import_policy_command.add_argument("-d", "--description", help="Description of new policy", required=True)
    import_policy_command.add_argument("-p", "--prioritylevel", help="Priority level (HIGH, MEDIUM, LOW)",
                                       default="LOW")
    import_policy_command.add_argument("-f", "--policyfile", help="Filename containing the JSON policy description",
                                       required=True)

    export_policy_command = commands.add_parser("export", help="Export policy to JSON file")
    export_policy_specifier = export_policy_command.add_mutually_exclusive_group(required=False)
    export_policy_specifier.add_argument("-i", "--id", type=int, help="ID of policy")
    export_policy_specifier.add_argument("-N", "--name", help="Name of policy")

    del_command = commands.add_parser("delete", help="Delete policies")
    del_policy_specifier = del_command.add_mutually_exclusive_group(required=True)
    del_policy_specifier.add_argument("-i", "--id", type=int, help="ID of policy to delete")
    del_policy_specifier.add_argument("-N", "--name", help="Name of policy to delete. Specify --force to delete"
                                      " multiple policies that have the same name")
    del_command.add_argument("--force", help="If NAME matches multiple policies, delete all matching policies",
                             action="store_true", default=False)

    add_rule_command = commands.add_parser("add-rule", help="Add rule to existing policy from JSON rule file")
    add_rule_specifier = add_rule_command.add_mutually_exclusive_group(required=True)
    add_rule_specifier.add_argument("-i", "--id", type=int, help="ID of policy (can specify multiple)",
                                    action="append", metavar="POLICYID")
    add_rule_specifier.add_argument("-N", "--name", help="Name of policy (can specify multiple)",
                                    action="append", metavar="POLICYNAME")
    add_rule_command.add_argument("-f", "--rulefile", help="Filename containing the JSON rule", required=True)

    del_rule_command = commands.add_parser("del-rule", help="Delete rule from existing policy")
    del_rule_specifier = del_rule_command.add_mutually_exclusive_group(required=True)
    del_rule_specifier.add_argument("-i", "--id", type=int, help="ID of policy")
    del_rule_specifier.add_argument("-N", "--name", help="Name of policy")
    del_rule_command.add_argument("-r", "--ruleid", type=int, help="ID of rule", required=True)

    replace_rule_command = commands.add_parser("replace-rule", help="Replace existing rule with a new one")
    replace_rule_specifier = replace_rule_command.add_mutually_exclusive_group(required=True)
    replace_rule_specifier.add_argument("-i", "--id", type=int, help="ID of policy")
    replace_rule_specifier.add_argument("-N", "--name", help="Name of policy")
    replace_rule_command.add_argument("-r", "--ruleid", type=int, help="ID of rule", required=True)
    replace_rule_command.add_argument("-f", "--rulefile", help="Filename containing the JSON rule", required=True)

    args = parser.parse_args()
    cb = get_cb_defense_object(args)

    if args.command_name == "list":
        return list_policies(cb, parser, args)
    elif args.command_name == "import":
        return import_policy(cb, parser, args)
    elif args.command_name == "export":
        return export_policy(cb, parser, args)
    elif args.command_name == "delete":
        return delete_policy(cb, parser, args)
    elif args.command_name == "add-rule":
        return add_rule(cb, parser, args)
    elif args.command_name == "del-rule":
        return del_rule(cb, parser, args)
    elif args.command_name == "replace-rule":
        return replace_rule(cb, parser, args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import sys

from cbapi.errors import ObjectNotFoundError

from cbapi.response.models import Process
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import csv
from cbapi.six import PY3
from cbapi.response.models import CbChildProcEvent, CbFileModEvent, CbNetConnEvent, CbRegModEvent, \
                                  CbModLoadEvent, CbCrossProcEvent


# UnicodeWriter class from http://python3porting.com/problems.html
class UnicodeWriter:
    def __init__(self, filename, dialect=csv.excel,
                 encoding="utf-8", **kw):
        self.filename = filename
        self.dialect = dialect
        self.encoding = encoding
        self.kw = kw

    def __enter__(self):
        if PY3:
            self.f = open(self.filename, 'wt',
                          encoding=self.encoding, newline='')
        else:
            self.f = open(self.filename, 'wb')
        self.writer = csv.writer(self.f, dialect=self.dialect,
                                 **self.kw)
        return self

    def __exit__(self, type, value, traceback):
        self.f.close()

    def writerow(self, row):
        if not PY3:
            row = [s or "" for s in row]
            row = [s.encode(self.encoding) for s in row]
        self.writer.writerow(row)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


def event_summary(event):
    timestamp = str(event.timestamp)
    if type(event) == CbFileModEvent:
        return [event.parent.path, timestamp, event.type, event.path, '']
    elif type(event) == CbNetConnEvent:
        if event.domain:
            hostname = event.domain
        else:
            hostname = event.remote_ip
        hostname += ':%d' % event.remote_port

        return [event.parent.path, timestamp, event.direction + ' netconn', hostname, '']
    elif type(event) == CbRegModEvent:
        return [event.parent.path, timestamp, event.type, event.path, '']
    elif type(event) == CbChildProcEvent:
        try:
            childproc = event.process.cmdline
        except ObjectNotFoundError:
            childproc = "<unknown>"
        return [event.parent.path, timestamp, 'childproc', event.path, childproc]
    elif type(event) == CbModLoadEvent:
        return [event.parent.path, timestamp, 'modload', event.path, event.md5]
    elif type(event) == CbCrossProcEvent:
        return [event.source_path, timestamp, event.type, event.target_path, event.privileges]
    else:
        return None


def write_csv(proc, filename):
    total_events = 0
    written_events = 0

    with UnicodeWriter(filename) as eventwriter:
        eventwriter.writerow(['ProcessPath', 'Timestamp', 'Event', 'Path/IP/Domain', 'Comments'])
        for event in proc.all_events:
            total_events += 1
            summary = event_summary(event)
            if summary:
                written_events += 1
                eventwriter.writerow(summary)

    print("{0} events out of {1} total events exported from process at {2}".format(written_events, total_events,
                                                                                   proc.webui_link))


def main():
    parser = build_cli_parser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--processid", help="Process ID or URL to Process Analysis page")
    group.add_argument("--query", help="query to pull multiple processes")
    args = parser.parse_args()

    cb = get_cb_response_object(args)

    if args.processid:
        if args.processid.startswith("http"):
            # interpret as a URL
            proc = cb.from_ui(args.processid)
        else:
            # interpret as a Process ID
            proc = cb.select(Process, args.processid)

        write_csv(proc, "{0}.csv".format(proc.id))
    else:
        for proc in cb.select(Process).where(args.query).group_by("id"):
            write_csv(proc, "{0}.csv".format(proc.id))


if __name__ == "__main__":
    sys.exit(main())



from cbapi.errors import ApiError
from .base_query import PSCQueryBase, QueryBuilder, QueryBuilderSupportMixin, IterableQueryMixin
from .devices_query import DeviceSearchQuery


class BaseAlertSearchQuery(PSCQueryBase, QueryBuilderSupportMixin, IterableQueryMixin):
    """
    Represents a query that is used to locate BaseAlert objects.
    """
    VALID_CATEGORIES = ["THREAT", "MONITORED", "INFO", "MINOR", "SERIOUS", "CRITICAL"]
    VALID_REPUTATIONS = ["KNOWN_MALWARE", "SUSPECT_MALWARE", "PUP", "NOT_LISTED", "ADAPTIVE_WHITE_LIST",
                         "COMMON_WHITE_LIST", "TRUSTED_WHITE_LIST", "COMPANY_BLACK_LIST"]
    VALID_ALERT_TYPES = ["CB_ANALYTICS", "VMWARE", "WATCHLIST"]
    VALID_WORKFLOW_VALS = ["OPEN", "DISMISSED"]
    VALID_FACET_FIELDS = ["ALERT_TYPE", "CATEGORY", "REPUTATION", "WORKFLOW", "TAG", "POLICY_ID",
                          "POLICY_NAME", "DEVICE_ID", "DEVICE_NAME", "APPLICATION_HASH",
                          "APPLICATION_NAME", "STATUS", "RUN_STATE", "POLICY_APPLIED_STATE",
                          "POLICY_APPLIED", "SENSOR_ACTION"]

    def __init__(self, doc_class, cb):
        super().__init__(doc_class, cb)
        self._query_builder = QueryBuilder()
        self._criteria = {}
        self._time_filter = {}
        self._sortcriteria = {}
        self._bulkupdate_url = "/appservices/v6/orgs/{0}/alerts/workflow/_criteria"

    def _update_criteria(self, key, newlist):
        """
        Updates the criteria being collected for a query. Assumes the specified criteria item is
        defined as a list; the list passed in will be set as the value for this criteria item, or
        appended to the existing one if there is one.

        :param str key: The key for the criteria item to be set
        :param list newlist: List of values to be set for the criteria item
        """
        oldlist = self._criteria.get(key, [])
        self._criteria[key] = oldlist + newlist

    def set_categories(self, categories):
        """
        Restricts the alerts that this query is performed on to the specified categories.

        :param categories list: List of categories to be restricted to. Valid categories are
                                "THREAT", "MONITORED", "INFO", "MINOR", "SERIOUS", and "CRITICAL."
        :return: This instance
        """
        if not all((c in BaseAlertSearchQuery.VALID_CATEGORIES) for c in categories):
            raise ApiError("One or more invalid category values")
        self._update_criteria("category", categories)
        return self

    def set_create_time(self, *args, **kwargs):
        """
        Restricts the alerts that this query is performed on to the specified
        creation time (either specified as a start and end point or as a
        range).

        :return: This instance
        """
        if kwargs.get("start", None) and kwargs.get("end", None):
            if kwargs.get("range", None):
                raise ApiError("cannot specify range= in addition to start= and end=")
            stime = kwargs["start"]
            if not isinstance(stime, str):
                stime = stime.isoformat()
            etime = kwargs["end"]
            if not isinstance(etime, str):
                etime = etime.isoformat()
            self._time_filter = {"start": stime, "end": etime}
        elif kwargs.get("range", None):
            if kwargs.get("start", None) or kwargs.get("end", None):
                raise ApiError("cannot specify start= or end= in addition to range=")
            self._time_filter = {"range": kwargs["range"]}
        else:
            raise ApiError("must specify either start= and end= or range=")
        return self

    def set_device_ids(self, device_ids):
        """
        Restricts the alerts that this query is performed on to the specified
        device IDs.

        :param device_ids list: list of integer device IDs
        :return: This instance
        """
        if not all(isinstance(device_id, int) for device_id in device_ids):
            raise ApiError("One or more invalid device IDs")
        self._update_criteria("device_id", device_ids)
        return self

    def set_device_names(self, device_names):
        """
        Restricts the alerts that this query is performed on to the specified
        device names.

        :param device_names list: list of string device names
        :return: This instance
        """
        if not all(isinstance(n, str) for n in device_names):
            raise ApiError("One or more invalid device names")
        self._update_criteria("device_name", device_names)
        return self

    def set_device_os(self, device_os):
        """
        Restricts the alerts that this query is performed on to the specified
        device operating systems.

        :param device_os list: List of string operating systems.  Valid values are
                               "WINDOWS", "ANDROID", "MAC", "IOS", "LINUX", and "OTHER."
        :return: This instance
        """
        if not all((osval in DeviceSearchQuery.VALID_OS) for osval in device_os):
            raise ApiError("One or more invalid operating systems")
        self._update_criteria("device_os", device_os)
        return self

    def set_device_os_versions(self, device_os_versions):
        """
        Restricts the alerts that this query is performed on to the specified
        device operating system versions.

        :param device_os_versions list: List of string operating system versions.
        :return: This instance
        """
        if not all(isinstance(n, str) for n in device_os_versions):
            raise ApiError("One or more invalid device OS versions")
        self._update_criteria("device_os_version", device_os_versions)
        return self

    def set_device_username(self, users):
        """
        Restricts the alerts that this query is performed on to the specified
        user names.

        :param users list: List of string user names.
        :return: This instance
        """
        if not all(isinstance(u, str) for u in users):
            raise ApiError("One or more invalid user names")
        self._update_criteria("device_username", users)
        return self

    def set_group_results(self, do_group):
        """
        Specifies whether or not to group the results of the query.

        :param do_group boolean: True to group the results, False to not do so.
        :return: This instance
        """
        self._criteria["group_results"] = True if do_group else False
        return self

    def set_alert_ids(self, alert_ids):
        """
        Restricts the alerts that this query is performed on to the specified
        alert IDs.

        :param alert_ids list: List of string alert IDs.
        :return: This instance
        """
        if not all(isinstance(v, str) for v in alert_ids):
            raise ApiError("One or more invalid alert ID values")
        self._update_criteria("id", alert_ids)
        return self

    def set_legacy_alert_ids(self, alert_ids):
        """
        Restricts the alerts that this query is performed on to the specified
        legacy alert IDs.

        :param alert_ids list: List of string legacy alert IDs.
        :return: This instance
        """
        if not all(isinstance(v, str) for v in alert_ids):
            raise ApiError("One or more invalid alert ID values")
        self._update_criteria("legacy_alert_id", alert_ids)
        return self

    def set_minimum_severity(self, severity):
        """
        Restricts the alerts that this query is performed on to the specified
        minimum severity level.

        :param severity int: The minimum severity level for alerts.
        :return: This instance
        """
        self._criteria["minimum_severity"] = severity
        return self

    def set_policy_ids(self, policy_ids):
        """
        Restricts the alerts that this query is performed on to the specified
        policy IDs.

        :param policy_ids list: list of integer policy IDs
        :return: This instance
        """
        if not all(isinstance(policy_id, int) for policy_id in policy_ids):
            raise ApiError("One or more invalid policy IDs")
        self._update_criteria("policy_id", policy_ids)
        return self

    def set_policy_names(self, policy_names):
        """
        Restricts the alerts that this query is performed on to the specified
        policy names.

        :param policy_names list: list of string policy names
        :return: This instance
        """
        if not all(isinstance(n, str) for n in policy_names):
            raise ApiError("One or more invalid policy names")
        self._update_criteria("policy_name", policy_names)
        return self

    def set_process_names(self, process_names):
        """
        Restricts the alerts that this query is performed on to the specified
        process names.

        :param process_names list: list of string process names
        :return: This instance
        """
        if not all(isinstance(n, str) for n in process_names):
            raise ApiError("One or more invalid process names")
        self._update_criteria("process_name", process_names)
        return self

    def set_process_sha256(self, shas):
        """
        Restricts the alerts that this query is performed on to the specified
        process SHA-256 hash values.

        :param shas list: list of string process SHA-256 hash values
        :return: This instance
        """
        if not all(isinstance(n, str) for n in shas):
            raise ApiError("One or more invalid SHA256 values")
        self._update_criteria("process_sha256", shas)
        return self

    def set_reputations(self, reps):
        """
        Restricts the alerts that this query is performed on to the specified
        reputation values.

        :param reps list: List of string reputation values.  Valid values are
                          "KNOWN_MALWARE", "SUSPECT_MALWARE", "PUP", "NOT_LISTED",
                          "ADAPTIVE_WHITE_LIST", "COMMON_WHITE_LIST",
                          "TRUSTED_WHITE_LIST", and "COMPANY_BLACK_LIST".
        :return: This instance
        """
        if not all((r in BaseAlertSearchQuery.VALID_REPUTATIONS) for r in reps):
            raise ApiError("One or more invalid reputation values")
        self._update_criteria("reputation", reps)
        return self

    def set_tags(self, tags):
        """
        Restricts the alerts that this query is performed on to the specified
        tag values.

        :param tags list: list of string tag values
        :return: This instance
        """
        if not all(isinstance(tag, str) for tag in tags):
            raise ApiError("One or more invalid tags")
        self._update_criteria("tag", tags)
        return self

    def set_target_priorities(self, priorities):
        """
        Restricts the alerts that this query is performed on to the specified
        target priority values.

        :param priorities list: List of string target priority values.  Valid values are
                                "LOW", "MEDIUM", "HIGH", and "MISSION_CRITICAL".
        :return: This instance
        """
        if not all((prio in DeviceSearchQuery.VALID_PRIORITIES) for prio in priorities):
            raise ApiError("One or more invalid priority values")
        self._update_criteria("target_value", priorities)
        return self

    def set_threat_ids(self, threats):
        """
        Restricts the alerts that this query is performed on to the specified
        threat ID values.

        :param threats list: list of string threat ID values
        :return: This instance
        """
        if not all(isinstance(t, str) for t in threats):
            raise ApiError("One or more invalid threat ID values")
        self._update_criteria("threat_id", threats)
        return self

    def set_types(self, alerttypes):
        """
        Restricts the alerts that this query is performed on to the specified
        alert type values.

        :param alerttypes list: List of string alert type values.  Valid values are
                                "CB_ANALYTICS", "VMWARE", and "WATCHLIST".
        :return: This instance
        """
        if not all((t in BaseAlertSearchQuery.VALID_ALERT_TYPES) for t in alerttypes):
            raise ApiError("One or more invalid alert type values")
        self._update_criteria("type", alerttypes)
        return self

    def set_workflows(self, workflow_vals):
        """
        Restricts the alerts that this query is performed on to the specified
        workflow status values.

        :param workflow_vals list: List of string alert type values.  Valid values are
                                   "OPEN" and "DISMISSED".
        :return: This instance
        """
        if not all((t in BaseAlertSearchQuery.VALID_WORKFLOW_VALS) for t in workflow_vals):
            raise ApiError("One or more invalid workflow status values")
        self._update_criteria("workflow", workflow_vals)
        return self

    def _build_criteria(self):
        """
        Builds the criteria object for use in a query.

        :return: The criteria object.
        """
        mycrit = self._criteria
        if self._time_filter:
            mycrit["create_time"] = self._time_filter
        return mycrit

    def sort_by(self, key, direction="ASC"):
        """Sets the sorting behavior on a query's results.

        Example::

        >>> cb.select(BaseAlert).sort_by("name")

        :param key: the key in the schema to sort by
        :param direction: the sort order, either "ASC" or "DESC"
        :rtype: :py:class:`BaseAlertSearchQuery`
        """
        if direction not in DeviceSearchQuery.VALID_DIRECTIONS:
            raise ApiError("invalid sort direction specified")
        self._sortcriteria = {"field": key, "order": direction}
        return self

    def _build_request(self, from_row, max_rows, add_sort=True):
        """
        Creates the request body for an API call.

        :param int from_row: The row to start the query at.
        :param int max_rows: The maximum number of rows to be returned.
        :param boolean add_sort: If True(default), the sort criteria will be added as part of the request.
        :return: A dict containing the complete request body.
        """
        request = {"criteria": self._build_criteria()}
        request["query"] = self._query_builder._collapse()
        if from_row > 0:
            request["start"] = from_row
        if max_rows >= 0:
            request["rows"] = max_rows
        if add_sort and self._sortcriteria != {}:
            request["sort"] = [self._sortcriteria]
        return request

    def _build_url(self, tail_end):
        """
        Creates the URL to be used for an API call.

        :param str tail_end: String to be appended to the end of the generated URL.
        """
        url = self._doc_class.urlobject.format(self._cb.credentials.org_key) + tail_end
        return url

    def _count(self):
        """
        Returns the number of results from the run of this query.

        :return: The number of results from the run of this query.
        """
        if self._count_valid:
            return self._total_results

        url = self._build_url("/_search")
        request = self._build_request(0, -1)
        resp = self._cb.post_object(url, body=request)
        result = resp.json()

        self._total_results = result["num_found"]
        self._count_valid = True

        return self._total_results

    def _perform_query(self, from_row=0, max_rows=-1):
        """
        Performs the query and returns the results of the query in an iterable fashion.

        :param int from_row: The row to start the query at (default 0).
        :param int max_rows: The maximum number of rows to be returned (default -1, meaning "all").
        """
        url = self._build_url("/_search")
        current = from_row
        numrows = 0
        still_querying = True
        while still_querying:
            request = self._build_request(current, max_rows)
            resp = self._cb.post_object(url, body=request)
            result = resp.json()

            self._total_results = result["num_found"]
            self._count_valid = True

            results = result.get("results", [])
            for item in results:
                yield self._doc_class(self._cb, item["id"], item)
                current += 1
                numrows += 1

                if max_rows > 0 and numrows == max_rows:
                    still_querying = False
                    break

            from_row = current
            if current >= self._total_results:
                still_querying = False
                break

    def facets(self, fieldlist, max_rows=0):
        """
        Return information about the facets for this alert by search, using the defined criteria.

        :param fieldlist list: List of facet field names. Valid names are
                               "ALERT_TYPE", "CATEGORY", "REPUTATION", "WORKFLOW", "TAG", "POLICY_ID",
                               "POLICY_NAME", "DEVICE_ID", "DEVICE_NAME", "APPLICATION_HASH",
                               "APPLICATION_NAME", "STATUS", "RUN_STATE", "POLICY_APPLIED_STATE",
                               "POLICY_APPLIED", and "SENSOR_ACTION".
        :param max_rows int: The maximum number of rows to return. 0 means return all rows.
        :return: A list of facet information specified as dicts.
        """
        if not all((field in BaseAlertSearchQuery.VALID_FACET_FIELDS) for field in fieldlist):
            raise ApiError("One or more invalid term field names")
        request = self._build_request(0, -1, False)
        request["terms"] = {"fields": fieldlist, "rows": max_rows}
        url = self._build_url("/_facet")
        resp = self._cb.post_object(url, body=request)
        result = resp.json()
        return result.get("results", [])

    def _update_status(self, status, remediation, comment):
        """
        Updates the status of all alerts matching the given query.

        :param str state: The status to put the alerts into, either "OPEN" or "DISMISSED".
        :param remediation str: The remediation state to set for all alerts.
        :param comment str: The comment to set for all alerts.
        :return: The request ID, which may be used to select a WorkflowStatus object.
        """
        request = {"state": status, "criteria": self._build_criteria(), "query": self._query_builder._collapse()}
        if remediation is not None:
            request["remediation_state"] = remediation
        if comment is not None:
            request["comment"] = comment
        resp = self._cb.post_object(self._bulkupdate_url.format(self._cb.credentials.org_key), body=request)
        output = resp.json()
        return output["request_id"]

    def update(self, remediation=None, comment=None):
        """
        Update all alerts matching the given query. The alerts will be left in an OPEN state after this request.

        :param remediation str: The remediation state to set for all alerts.
        :param comment str: The comment to set for all alerts.
        :return: The request ID, which may be used to select a WorkflowStatus object.
        """
        return self._update_status("OPEN", remediation, comment)

    def dismiss(self, remediation=None, comment=None):
        """
        Dismiss all alerts matching the given query. The alerts will be left in a DISMISSED state after this request.

        :param remediation str: The remediation state to set for all alerts.
        :param comment str: The comment to set for all alerts.
        :return: The request ID, which may be used to select a WorkflowStatus object.
        """
        return self._update_status("DISMISSED", remediation, comment)


class WatchlistAlertSearchQuery(BaseAlertSearchQuery):
    """
    Represents a query that is used to locate WatchlistAlert objects.
    """
    def __init__(self, doc_class, cb):
        super().__init__(doc_class, cb)
        self._bulkupdate_url = "/appservices/v6/orgs/{0}/alerts/watchlist/workflow/_criteria"

    def set_watchlist_ids(self, ids):
        """
        Restricts the alerts that this query is performed on to the specified
        watchlist ID values.

        :param ids list: list of string watchlist ID values
        :return: This instance
        """
        if not all(isinstance(t, str) for t in ids):
            raise ApiError("One or more invalid watchlist IDs")
        self._update_criteria("watchlist_id", ids)
        return self

    def set_watchlist_names(self, names):
        """
        Restricts the alerts that this query is performed on to the specified
        watchlist name values.

        :param names list: list of string watchlist name values
        :return: This instance
        """
        if not all(isinstance(name, str) for name in names):
            raise ApiError("One or more invalid watchlist names")
        self._update_criteria("watchlist_name", names)
        return self


class CBAnalyticsAlertSearchQuery(BaseAlertSearchQuery):
    """
    Represents a query that is used to locate CBAnalyticsAlert objects.
    """
    VALID_THREAT_CATEGORIES = ["UNKNOWN", "NON_MALWARE", "NEW_MALWARE", "KNOWN_MALWARE", "RISKY_PROGRAM"]
    VALID_LOCATIONS = ["ONSITE", "OFFSITE", "UNKNOWN"]
    VALID_KILL_CHAIN_STATUSES = ["RECONNAISSANCE", "WEAPONIZE", "DELIVER_EXPLOIT", "INSTALL_RUN",
                                 "COMMAND_AND_CONTROL", "EXECUTE_GOAL", "BREACH"]
    VALID_POLICY_APPLIED = ["APPLIED", "NOT_APPLIED"]
    VALID_RUN_STATES = ["DID_NOT_RUN", "RAN", "UNKNOWN"]
    VALID_SENSOR_ACTIONS = ["POLICY_NOT_APPLIED", "ALLOW", "ALLOW_AND_LOG", "TERMINATE", "DENY"]
    VALID_THREAT_CAUSE_VECTORS = ["EMAIL", "WEB", "GENERIC_SERVER", "GENERIC_CLIENT", "REMOTE_DRIVE",
                                  "REMOVABLE_MEDIA", "UNKNOWN", "APP_STORE", "THIRD_PARTY"]

    def __init__(self, doc_class, cb):
        super().__init__(doc_class, cb)
        self._bulkupdate_url = "/appservices/v6/orgs/{0}/alerts/cbanalytics/workflow/_criteria"

    def set_blocked_threat_categories(self, categories):
        """
        Restricts the alerts that this query is performed on to the specified
        threat categories that were blocked.

        :param categories list: List of threat categories to look for.  Valid values are "UNKNOWN",
                                "NON_MALWARE", "NEW_MALWARE", "KNOWN_MALWARE", and "RISKY_PROGRAM".
        :return: This instance.
        """
        if not all((category in CBAnalyticsAlertSearchQuery.VALID_THREAT_CATEGORIES)
                   for category in categories):
            raise ApiError("One or more invalid threat categories")
        self._update_criteria("blocked_threat_category", categories)
        return self

    def set_device_locations(self, locations):
        """
        Restricts the alerts that this query is performed on to the specified
        device locations.

        :param locations list: List of device locations to look for. Valid values are "ONSITE", "OFFSITE",
                               and "UNKNOWN".
        :return: This instance.
        """
        if not all((location in CBAnalyticsAlertSearchQuery.VALID_LOCATIONS)
                   for location in locations):
            raise ApiError("One or more invalid device locations")
        self._update_criteria("device_location", locations)
        return self

    def set_kill_chain_statuses(self, statuses):
        """
        Restricts the alerts that this query is performed on to the specified
        kill chain statuses.

        :param statuses list: List of kill chain statuses to look for. Valid values are "RECONNAISSANCE",
                              "WEAPONIZE", "DELIVER_EXPLOIT", "INSTALL_RUN","COMMAND_AND_CONTROL",
                              "EXECUTE_GOAL", and "BREACH".
        :return: This instance.
        """
        if not all((status in CBAnalyticsAlertSearchQuery.VALID_KILL_CHAIN_STATUSES)
                   for status in statuses):
            raise ApiError("One or more invalid kill chain status values")
        self._update_criteria("kill_chain_status", statuses)
        return self

    def set_not_blocked_threat_categories(self, categories):
        """
        Restricts the alerts that this query is performed on to the specified
        threat categories that were NOT blocked.

        :param categories list: List of threat categories to look for.  Valid values are "UNKNOWN",
                                "NON_MALWARE", "NEW_MALWARE", "KNOWN_MALWARE", and "RISKY_PROGRAM".
        :return: This instance.
        """
        if not all((category in CBAnalyticsAlertSearchQuery.VALID_THREAT_CATEGORIES)
                   for category in categories):
            raise ApiError("One or more invalid threat categories")
        self._update_criteria("not_blocked_threat_category", categories)
        return self

    def set_policy_applied(self, applied_statuses):
        """
        Restricts the alerts that this query is performed on to the specified
        status values showing whether policies were applied.

        :param applied_statuses list: List of status values to look for. Valid values are
                                      "APPLIED" and "NOT_APPLIED".
        :return: This instance.
        """
        if not all((s in CBAnalyticsAlertSearchQuery.VALID_POLICY_APPLIED)
                   for s in applied_statuses):
            raise ApiError("One or more invalid policy-applied values")
        self._update_criteria("policy_applied", applied_statuses)
        return self

    def set_reason_code(self, reason):
        """
        Restricts the alerts that this query is performed on to the specified
        reason codes (enum values).

        :param reason list: List of string reason codes to look for.
        :return: This instance.
        """
        if not all(isinstance(t, str) for t in reason):
            raise ApiError("One or more invalid reason code values")
        self._update_criteria("reason_code", reason)
        return self

    def set_run_states(self, states):
        """
        Restricts the alerts that this query is performed on to the specified run states.

        :param states list: List of run states to look for. Valid values are "DID_NOT_RUN", "RAN",
                            and "UNKNOWN".
        :return: This instance.
        """
        if not all((s in CBAnalyticsAlertSearchQuery.VALID_RUN_STATES)
                   for s in states):
            raise ApiError("One or more invalid run states")
        self._update_criteria("run_state", states)
        return self

    def set_sensor_actions(self, actions):
        """
        Restricts the alerts that this query is performed on to the specified sensor actions.

        :param actions list: List of sensor actions to look for. Valid values are "POLICY_NOT_APPLIED",
                             "ALLOW", "ALLOW_AND_LOG", "TERMINATE", and "DENY".
        :return: This instance.
        """
        if not all((action in CBAnalyticsAlertSearchQuery.VALID_SENSOR_ACTIONS)
                   for action in actions):
            raise ApiError("One or more invalid sensor actions")
        self._update_criteria("sensor_action", actions)
        return self

    def set_threat_cause_vectors(self, vectors):
        """
        Restricts the alerts that this query is performed on to the specified threat cause vectors.

        :param vectors list: List of threat cause vectors to look for.  Valid values are "EMAIL", "WEB",
                             "GENERIC_SERVER", "GENERIC_CLIENT", "REMOTE_DRIVE", "REMOVABLE_MEDIA",
                             "UNKNOWN", "APP_STORE", and "THIRD_PARTY".
        :return: This instance.
        """
        if not all((vector in CBAnalyticsAlertSearchQuery.VALID_THREAT_CAUSE_VECTORS)
                   for vector in vectors):
            raise ApiError("One or more invalid threat cause vectors")
        self._update_criteria("threat_cause_vector", vectors)
        return self


class VMwareAlertSearchQuery(BaseAlertSearchQuery):
    """
    Represents a query that is used to locate VMwareAlert objects.
    """
    def __init__(self, doc_class, cb):
        super().__init__(doc_class, cb)
        self._bulkupdate_url = "/appservices/v6/orgs/{0}/alerts/vmware/workflow/_criteria"

    def set_group_ids(self, groupids):
        """
        Restricts the alerts that this query is performed on to the specified
        AppDefense-assigned alarm group IDs.

        :param groupids list: List of (integer) AppDefense-assigned alarm group IDs.
        :return: This instance.
        """
        if not all(isinstance(groupid, int) for groupid in groupids):
            raise ApiError("One or more invalid alarm group IDs")
        self._update_criteria("group_id", groupids)
        return self
